/*
 * Bresenham circle drawing functions with squeeze, filling and
 * clipping. Includes some helper functions like pset, clipped
 * pset, cls, progress bar. All for MCGA. Includes a comprehensive
 * test suite in main(). By Eric Auer, <eric@coli.uni-sb.de>, 2003.
 * Should be especially useful for "GW Basic style" circle drawing.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should find a copy of the GNU General Public License on the
 * site where you have downloaded this source. If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307 USA
 * (or try http://www.gnu.org/licenses/licenses.html at www.gnu.org).
 */

#include <dos.h>    /* MK_FP, int86, REGS, sleep */
#include <stdlib.h> /* random, randomize, srand */


int psetclipped(int, int, int);
void brecirc(int, int, int, int, double);
void brecircslow(int, int, int, int, double, int);
void xsleep(int);
void xscale(int, int, int);
void mcgacls(int);
int main(int, char **);


/*
 * MCGA simply uses a memory mapped pixel buffer
 */
unsigned char far * mcga = (unsigned char far *)MK_FP(0xa000, 0);


/*
 * Paint a pixel of specified color at specified point. No clipping.
 */
#define PSET(x,y,c) mcga[(x)+(320*(y))] = (c)


/*
 * Paint a pixel, if visible, of specified color at specified point.
 */
int psetclipped(int x, int y, int color)
{
  if ((x<0) || (x>=320) || (y<0) || (y>=200))
    return 1; /* clipped */
  PSET(x, y, color);
  return 0; /* not clipped */
}


/*
 * Circle drawing using the Bresenham algorithm
 * xm and ym are the center coordinates, r the radius,
 * and squeeze is the x / y radius ratio. Calls
 * brecircslow() if squeeze is not 1 or if clipping is needed.
 */
void brecirc(int xm, int ym, int r, int color, double squeeze)
{
  double factor;
  int d; /* diagonal */
  int x, y; /* coordinate in one octet */
  int maxr1, maxr2; /* clipping check */
  maxr1 = (xm < 160) ? xm : (320-xm);
  maxr2 = (ym < 100) ? ym : (200-ym);
  d = 3 - (r + r);
  x = 0;
  y = r;
  factor = (squeeze > 1.0) ? (1.0 / squeeze) : squeeze;
  if ((factor < 0.99) || (r > maxr1) || (r > maxr2)) {
    brecircslow(xm, ym, r, color, squeeze, 0);
    return;
  }
  /* loop to draw one octet only, from 0 to 45 degrees: */
  while (x <= y) {
    /* the octets close to the y axis: */
    PSET(xm+x, ym+y, color);
    PSET(xm-x, ym+y, color);
    PSET(xm+x, ym-y, color);
    PSET(xm-x, ym-y, color);
    /* the octets close to the x axis swap variables: */
    PSET(xm+y, ym+x, color);
    PSET(xm-y, ym+x, color);
    PSET(xm+y, ym-x, color);
    PSET(xm-y, ym-x, color);
    /* now go one step to the right or diagonal: */
    x++;
    if (d < 0) {
      d += (x << 2); /* (x+1)*4 + 2 */
    } else {
      y--;
      d += ((x-y) << 2); /* (2+x-y)*4 + 2 or 4*(x-y) + 10 */
    }
    d++;
    d++;
  } /* while */
} /* brecirc */


/*
 * Circle drawing using the Bresenham algorithm
 * xm and ym are the center coordinates, r the radius,
 * and squeeze IS the x / y radius ratio.
 * This version is slower and does clipping.
 * It also optionally fills the circle.
 */
void brecircslow(int xm, int ym, int r, int color, double squeeze, int fill)
{
  double factor;
  int d;    /* diagonal */
  int x, y; /* coordinate in one octet */
  int i;    /* counter for filling */
  double x2f, y2f; /* squeezed */
  int x2, y2;      /* squeezed */
  d = 3 - (r + r);
  factor = (squeeze > 1.0) ? (1.0 / squeeze) : squeeze;

  if (squeeze > 1.0) { /* shorten Y dimension */

    x = 0;
    x2f = 0;
    x2 = 0;
    y = r;
    y2f = y * factor;
    y2 = (int) y2f;
    /* loop to draw one octet only, from 0 to 45 degrees: */
    while (x <= y) {
      if (fill == 0) {
        /* the octets close to the y axis: */
        psetclipped(xm+x, ym+y2, color);
        psetclipped(xm-x, ym+y2, color);
        psetclipped(xm+x, ym-y2, color);
        psetclipped(xm-x, ym-y2, color);
        /* the octets close to the x axis swap variables: */
        psetclipped(xm+y, ym+x2, color);
        psetclipped(xm-y, ym+x2, color);
        psetclipped(xm+y, ym-x2, color);
        psetclipped(xm-y, ym-x2, color);
      } else { /* filled */
        for (i=-y2; i<y2; i++) {
          psetclipped(xm+x, ym+i, color);
          psetclipped(xm-x, ym+i, color);
        }
        for (i=-x2; i<x2; i++) {
          psetclipped(xm+y, ym+i, color);
          psetclipped(xm-y, ym+i, color);
        }
      } /* filled */
      /* now go one step to the right or diagonal: */
      x++;
      x2f = x2f + factor;
      x2 = (int) x2f;
      if (d < 0) {
        d += (x << 2); /* (x+1)*4 + 2 */
      } else {
        y--;
        y2f = y2f - factor;
        y2 = (int) y2f;
        d += ((x-y) << 2); /* (2+x-y)*4 + 2 or 4*(x-y) + 10 */
      }
      d++;
      d++;
    } /* while */
    
  } else { /* shorten X dimension */

    y = 0;
    y2f = 0;
    y2 = 0;
    x = r;
    x2f = x * factor;
    x2 = (int) x2f;
    /* loop to draw one octet only, from 90 to 45 degrees: */
    while (y <= x) {
      if (fill == 0) {
        /* the octets close to the y axis: */
        psetclipped(xm+x2, ym+y, color);
        psetclipped(xm-x2, ym+y, color);
        psetclipped(xm+x2, ym-y, color);
        psetclipped(xm-x2, ym-y, color);
        /* the octets close to the x axis swap variables: */
        psetclipped(xm+y2, ym+x, color);
        psetclipped(xm-y2, ym+x, color);
        psetclipped(xm+y2, ym-x, color);
        psetclipped(xm-y2, ym-x, color);
      } else { /* filled */
        for (i=-x2; i<x2; i++) {
          psetclipped(xm+i, ym+y, color);
          psetclipped(xm+i, ym-y, color);
        }
        for (i=-y2; i<y2; i++) {
          psetclipped(xm+i, ym+x, color);
          psetclipped(xm+i, ym-x, color);
        }
      } /* filled */
      /* now go one step down or diagonal: */
      y++;
      y2f = y2f + factor;
      y2 = (int) y2f;
      if (d < 0) {
        d += (y << 2); /* (y+1)*4 + 2 */
      } else {
        x--;
        x2f = x2f - factor;
        x2 = (int) x2f;
        d += ((y-x) << 2); /* (2+y-x)*4 + 2 or 4*(y-x) + 10 */
      }
      d++;
      d++;
    } /* while */
      
  }

} /* brecircslow */


/*
 * Sleep a while and show a progress indicator in the meantime
 */
void xsleep(int secs)
{
  int rest;
  for (rest=secs ; rest>0; rest--) {
    xscale(rest, secs, 0); /* draw progress indicator */
    sleep(1);
  }
  xscale(rest, secs, 1);   /* remove progress indicator */
} /* xsleep */


/*
 * Show a progress indicator
 */
void xscale(int used, int maximum, int remove)
{
  int fac = 5;
  int i;
  if (remove != 0) {
    for (i=0; i<maximum; i++) /* remove bar of dots */
      brecircslow(fac+(2*fac*i), 10, fac-2, 0, 1.0, 1);
  } else {
    for (i=used; i<maximum; i++) /* red bar of dots */
      brecircslow(fac+(2*fac*i), 10, fac-2, 12, 1.0, 1);
    for (i=0; i<used; i++)    /* green part of bar */
      brecircslow(fac+(2*fac*i), 10, fac-2, 10, 1.0, 1);
  }
} /* xscale */


/*
 *
 */
void mcgacls(int color)
{
  unsigned int i, max;
  max = 320 * 200;
  for (i=0; i<max; i++)
    mcga[i] = color;
} /* mcgacls */


/*
 * main() just displays a few test patterns
 */
int main(int argc, char ** argv)
{
  int i, j;
  int x, y, r1, r2, r;
  union REGS regs;

  srand(42 + (argv != NULL) + argc);
    /* randomize would need time.h */
  regs.x.ax = 0x0013; /* MCGA */
  int86(0x10, &regs, &regs);

  /* fill whole screen with a concentric pattern */
  for (i=0; i<190; i++) /* sqrt(160*160 + 100*100) is 188 */
    brecirc(160, 100, i, 32 + ((3+2) * 24) + (i % 24), 1.0);
  xsleep(1);

  /* draw round circles without clipping */
  for (i=0; i<50; i++) {
    x = 10 + random(300);
    y = 10 + random(180);
    r1 = (x < 160) ? x : (320-x);
    r2 = (y < 100) ? y : (200-y);
    r = (r1 < r2) ? r1 : r2;
    r = random(r);
    brecirc(x, y, r, 32 + (i % 24), 1.0);
  }
  xsleep(2);

  /* try clipping and ellipses */
  for (i=0; i<10; i++)
    brecirc(random(320), random(200), random(160), 32 + (i % 24),
      (random(100) > 50) ? (10.0 / (1+random(200))) : (0.1 * random(200)) );
  xsleep(2);

  /* try clipped round circles */
  for (i=0; i<(3*24); i++)
    brecircslow(random(640), random(400), random(320), 32 + 48 + (i % 24),
      1.0, (i < 24) ? 1 : 0);
  xsleep(2);

  /* try clipped round filled circles */
  for (i=0; i<24; i++)
    brecircslow(random(640), random(400), random(160), 32 + i, 1.0, 1);
  xsleep(2);

  /* try a few ellipses more, with clipping, some filled */
  for (j=0; j<16; j++) {
    mcgacls(16 + j);
    for (i=-8; i<(3*24); i++)
      brecircslow(random(320), random(200), random(160), 32 + i,
        (random(100) > 50)
          ? (2.0 / (2+random(10)))
          : (0.5 * (2+random(10))),
        (i < 0) ? 1 : 0 );
    xsleep(1);
  }

  /* benchmark for clipped and filled ellipses */
  mcgacls(0);
  for (j=0; j<20; j++) {
    for (i=0; i<100; i++) {
      brecirc(random(320), random(200), random(160), 32 + (3*24) + (i % 24),
        (random(100) > 50) ? (10.0 / (1+random(200))) : (0.1 * random(200)) );
      xscale(j, 20, 0); /* progress indicator */
    }
  }
  xscale(20, 20, 1);  /* remove progress indicator */
  xsleep(2);
  
  /* benchmark for filled ellipses */
  mcgacls(0);
  for (j=0; j<20; j++) {
    for (i=0; i<100; i++) {
      brecircslow(random(320), random(200), random(160),
        32 + (3*24) + (i % 24),
        (random(100) > 50)
          ? (10.0 / (1+random(200)))
          : (0.1 * random(200)),
        1 );
      xscale(j, 20, 0); /* progress indicator */
    }
  }
  xscale(20, 20, 1);  /* remove progress indicator */
  xsleep(2);
  
  regs.x.ax = 0x0003; /* TEXT */
  int86(0x10, &regs, &regs);

} /* main */
